import types
import pytest

from py_azpe import AzurePolicyEngine


class DummyAssignmentsClient:
    def __init__(self):
        self.created = []

    def create(self, scope, name, params):
        self.created.append((scope, name, params))
        return types.SimpleNamespace(
            as_dict=lambda: {
                "scope": scope,
                "name": name,
                "params_type": type(params).__name__,
            }
        )


class DummyPolicyClient:
    def __init__(self):
        self.policy_assignments = DummyAssignmentsClient()


def _patch_policy_client(monkeypatch, dummy):
    monkeypatch.setitem(
        __import__("sys").modules,
        "azure.mgmt.resource",
        types.SimpleNamespace(PolicyClient=lambda cred, sub: dummy),
    )


def test_prepare_assignment_payload_returns_expected(monkeypatch, tmp_path):
    engine = AzurePolicyEngine(backend="sdk", policy_dir=str(tmp_path))
    # no SDK present -> prepare should still produce payload
    assign = {
        "properties": {
            "displayName": "DRY",
            "policyDefinitionId": "/subscriptions/0000/providers/Microsoft.Authorization/policyDefinitions/pd1",
        }
    }
    res = engine.prepare_assignment_payload("dry1", assign, scope="/subscriptions/0000")
    assert res["name"] == "dry1"
    assert res["scope"] == "/subscriptions/0000"
    assert "payload" in res


def test_validate_scope_rejects_bad(monkeypatch):
    engine = AzurePolicyEngine(backend="file")
    with pytest.raises(ValueError):
        engine._validate_scope("not-a-scope")
    with pytest.raises(ValueError):
        engine._validate_scope("")


def test_dry_run_with_mgmt_group(monkeypatch, tmp_path):
    engine = AzurePolicyEngine(backend="sdk", policy_dir=str(tmp_path))
    dummy = DummyPolicyClient()
    _patch_policy_client(monkeypatch, dummy)

    assign = {
        "properties": {
            "displayName": "MG DRY",
            "policyDefinitionId": "/providers/Microsoft.Management/managementGroups/mg1/providers/Microsoft.Authorization/policyDefinitions/pd1",
        }
    }
    res = engine.prepare_assignment_payload(
        "mgDry", assign, scope="/providers/Microsoft.Management/managementGroups/mg1"
    )
    assert res["scope"].endswith("mg1")
    assert res["name"] == "mgDry"



"""Duplicate package copy (for importable package name `py_azpe`).
This file mirrors the implementation in the repository folder `azure-policy-engine/engine.py`.
"""
from __future__ import annotations

import importlib
import json
import os
from typing import Any, Dict, List, Optional, TYPE_CHECKING


# Optional type-only imports for static analysis
if TYPE_CHECKING:
    try:
        from azure.mgmt.resource import PolicyClient  # type: ignore
        import azure.mgmt.resource.policy as _azure_mgmt_policy  # type: ignore
    except Exception:  # pragma: no cover - only for static analysis
        PolicyClient = Any  # type: ignore
        _azure_mgmt_policy = Any  # type: ignore


def _import_attr(mod_name: str, attr: str):
    try:
        mod = importlib.import_module(mod_name)
        return getattr(mod, attr, None)
    except Exception:
        return None


CJSON_EXT = ".json"
_AZURE_MGMT_RESOURCE = "azure.mgmt.resource"
_AZURE_MGMT_POLICY_MODELS = "azure.mgmt.resource.policy.models"
_CREDENTIAL_ERR = "Azure credential not provided for SDK/ARM calls"


class AzurePolicyEngine:
    def __init__(
        self,
        backend: str = "file",
        policy_dir: Optional[str] = None,
        subscription_id: Optional[str] = None,
        credential: Optional[Any] = None,
    ) -> None:
        self.backend = backend
        base = os.path.dirname(__file__)
        self.policy_dir = policy_dir or os.path.join(base, "policies")
        # allow defaults from environment via azure_auth helpers
        self.subscription_id = subscription_id
        self.credential = credential
        if self.subscription_id is None or self.credential is None:
            try:
                # lazy import so we don't require azure packages for file/azcli backends
                from .azure_auth import get_subscription_id, get_credential

                if self.subscription_id is None:
                    try:
                        self.subscription_id = get_subscription_id()
                    except Exception:
                        # keep None if not available
                        pass
                if self.credential is None:
                    try:
                        self.credential = get_credential()
                    except Exception:
                        # keep None if not available
                        pass
            except Exception:
                # azure_auth not present or import failed; ignore and continue
                pass

        # canonical folders
        self.definitions_dir = os.path.join(self.policy_dir, "definitions")
        self.assignments_dir = os.path.join(self.policy_dir, "assignments")
        self.initiatives_dir = os.path.join(self.policy_dir, "initiatives")

        os.makedirs(self.definitions_dir, exist_ok=True)
        os.makedirs(self.assignments_dir, exist_ok=True)
        os.makedirs(self.initiatives_dir, exist_ok=True)
        # optional azcli backend helper (lazy import)
        self._azcli = None

    # -----------------
    # File backend
    # -----------------
    def _file_list_policies(self) -> List[Dict[str, Any]]:
        results: List[Dict[str, Any]] = []
        # read JSON files from the definitions subfolder
        for fname in sorted(os.listdir(self.definitions_dir)):
            if not fname.lower().endswith(CJSON_EXT):
                continue
            path = os.path.join(self.definitions_dir, fname)
            try:
                with open(path, "r", encoding="utf-8") as f:
                    body = json.load(f)
                name = body.get("name") or os.path.splitext(fname)[0]
                results.append(
                    {
                        "name": name,
                        "file": path,
                        "properties": body.get("properties", body),
                    }
                )
            except Exception as ex:
                # include an entry with the error so callers can see problematic files
                results.append({"file": path, "error": str(ex)})
        return results

    def _file_get_policy(self, name: str) -> Dict[str, Any]:
        # try exact filename then name.json
        candidates = [name, f"{name}{CJSON_EXT}"]
        for c in candidates:
            path = os.path.join(self.definitions_dir, c)
            if os.path.exists(path):
                with open(path, "r", encoding="utf-8") as f:
                    return json.load(f)
        # fallback: search files for matching 'name' property
        for fname in os.listdir(self.definitions_dir):
            if not fname.lower().endswith(CJSON_EXT):
                continue
            path = os.path.join(self.definitions_dir, fname)
            try:
                with open(path, "r", encoding="utf-8") as f:
                    body = json.load(f)
                if body.get("name") == name or os.path.splitext(fname)[0] == name:
                    return body
            except Exception:
                continue
        raise FileNotFoundError(f"policy '{name}' not found in {self.policy_dir}")

    def _file_deploy_policy(self, name: str, policy: Dict[str, Any]) -> Dict[str, Any]:
        path = os.path.join(self.definitions_dir, f"{name}{CJSON_EXT}")
        with open(path, "w", encoding="utf-8") as f:
            json.dump(policy, f, indent=2)
        return {"file": path, "status": "written"}

    def _file_list_assignments(self) -> List[Dict[str, Any]]:
        results: List[Dict[str, Any]] = []
        for fname in sorted(os.listdir(self.assignments_dir)):
            if not fname.lower().endswith(CJSON_EXT):
                continue
            path = os.path.join(self.assignments_dir, fname)
            try:
                with open(path, "r", encoding="utf-8") as f:
                    body = json.load(f)
                name = body.get("name") or os.path.splitext(fname)[0]
                results.append(
                    {
                        "name": name,
                        "file": path,
                        "properties": body.get("properties", body),
                    }
                )
            except Exception as ex:
                results.append({"file": path, "error": str(ex)})
        return results

    def _file_list_initiatives(self) -> List[Dict[str, Any]]:
        results: List[Dict[str, Any]] = []
        for fname in sorted(os.listdir(self.initiatives_dir)):
            if not fname.lower().endswith(CJSON_EXT):
                continue
            path = os.path.join(self.initiatives_dir, fname)
            try:
                with open(path, "r", encoding="utf-8") as f:
                    body = json.load(f)
                name = body.get("name") or os.path.splitext(fname)[0]
                results.append(
                    {
                        "name": name,
                        "file": path,
                        "properties": body.get("properties", body),
                    }
                )
            except Exception as ex:
                results.append({"file": path, "error": str(ex)})
        return results

    def _file_get_initiative(self, name: str) -> Dict[str, Any]:
        candidates = [name, f"{name}{CJSON_EXT}"]
        for c in candidates:
            path = os.path.join(self.initiatives_dir, c)
            if os.path.exists(path):
                with open(path, "r", encoding="utf-8") as f:
                    return json.load(f)
        for fname in os.listdir(self.initiatives_dir):
            if not fname.lower().endswith(CJSON_EXT):
                continue
            path = os.path.join(self.initiatives_dir, fname)
            try:
                with open(path, "r", encoding="utf-8") as f:
                    body = json.load(f)
                if body.get("name") == name or os.path.splitext(fname)[0] == name:
                    return body
            except Exception:
                continue
        raise FileNotFoundError(
            f"initiative '{name}' not found in {self.initiatives_dir}"
        )

    def _file_deploy_initiative(
        self, name: str, initiative: Dict[str, Any]
    ) -> Dict[str, Any]:
        path = os.path.join(self.initiatives_dir, f"{name}{CJSON_EXT}")
        with open(path, "w", encoding="utf-8") as f:
            json.dump(initiative, f, indent=2)
        return {"file": path, "status": "written"}

    # -----------------
    # Public API
    # -----------------
    def list_policies(self) -> List[Dict[str, Any]]:
        """List available policy definitions from the selected backend.

        Returns a list of dicts with at least 'name' and 'properties'.
        """
        if self.backend == "file":
            return self._file_list_policies()
        elif self.backend == "sdk":
            return self._sdk_list_policies()
        elif self.backend == "azcli":
            # lazy import to avoid requiring azcli module at package import time
            if self._azcli is None:
                try:
                    from .azcli_backend import AzCliBackend  # type: ignore

                    self._azcli = AzCliBackend(subscription_id=self.subscription_id)
                except Exception:
                    raise RuntimeError(
                        "azcli backend is not available (ensure 'az' is installed)"
                    )
            vals = self._azcli.list_policy_definitions()
            # normalize to expected list items with name/properties
            out = []
            for v in vals:
                name = v.get("name") or v.get("properties", {}).get("displayName")
                out.append(
                    {"name": name, "file": None, "properties": v.get("properties", v)}
                )
            return out
        else:
            raise ValueError(f"unknown backend: {self.backend}")

    def get_policy(self, name: str) -> Dict[str, Any]:
        if self.backend == "file":
            return self._file_get_policy(name)
        elif self.backend == "sdk":
            return self._sdk_get_policy(name)
        elif self.backend == "azcli":
            if self._azcli is None:
                try:
                    from .azcli_backend import AzCliBackend  # type: ignore

                    self._azcli = AzCliBackend(subscription_id=self.subscription_id)
                except Exception:
                    raise RuntimeError(
                        "azcli backend is not available (ensure 'az' is installed)"
                    )
            return self._azcli.get_policy_definition(name)
        else:
            raise ValueError(f"unknown backend: {self.backend}")

    def deploy_policy(self, name: str, policy: Dict[str, Any]) -> Dict[str, Any]:
        """Create or update a policy definition with the given name and policy body."""
        if self.backend == "file":
            return self._file_deploy_policy(name, policy)
        elif self.backend == "sdk":
            return self._sdk_deploy_policy(name, policy)
        elif self.backend == "azcli":
            if self._azcli is None:
                try:
                    from .azcli_backend import AzCliBackend  # type: ignore

                    self._azcli = AzCliBackend(subscription_id=self.subscription_id)
                except Exception:
                    raise RuntimeError(
                        "azcli backend is not available (ensure 'az' is installed)"
                    )
            return self._azcli.create_or_update_policy_definition(name, policy)
        else:
            raise ValueError(f"unknown backend: {self.backend}")

    def list_assignments(self, scope: Optional[str] = None) -> List[Dict[str, Any]]:
        if self.backend == "file":
            return self._file_list_assignments()
        elif self.backend == "sdk":
            return self._sdk_list_assignments(scope)
        elif self.backend == "azcli":
            if self._azcli is None:
                try:
                    from .azcli_backend import AzCliBackend  # type: ignore

                    self._azcli = AzCliBackend(subscription_id=self.subscription_id)
                except Exception:
                    raise RuntimeError(
                        "azcli backend is not available (ensure 'az' is installed)"
                    )
            vals = self._azcli.list_assignments(scope)
            out = []
            for v in vals:
                out.append(
                    {
                        "name": v.get("name"),
                        "file": None,
                        "properties": v.get("properties", v),
                    }
                )
            return out
        else:
            raise ValueError(f"unknown backend: {self.backend}")

    def list_initiatives(self) -> List[Dict[str, Any]]:
        """List initiatives (policy set definitions)."""
        if self.backend == "file":
            return self._file_list_initiatives()
        elif self.backend == "sdk":
            return self._sdk_list_initiatives()
        elif self.backend == "azcli":
            if self._azcli is None:
                try:
                    from .azcli_backend import AzCliBackend  # type: ignore

                    self._azcli = AzCliBackend(subscription_id=self.subscription_id)
                except Exception:
                    raise RuntimeError(
                        "azcli backend is not available (ensure 'az' is installed)"
                    )
            vals = self._azcli.list_initiatives()
            out = []
            for v in vals:
                out.append(
                    {
                        "name": v.get("name"),
                        "file": None,
                        "properties": v.get("properties", v),
                    }
                )
            return out
        else:
            raise ValueError(f"unknown backend: {self.backend}")

    def get_initiative(self, name: str) -> Dict[str, Any]:
        if self.backend == "file":
            return self._file_get_initiative(name)
        elif self.backend == "sdk":
            return self._sdk_get_initiative(name)
        elif self.backend == "azcli":
            if self._azcli is None:
                try:
                    from .azcli_backend import AzCliBackend  # type: ignore

                    self._azcli = AzCliBackend(subscription_id=self.subscription_id)
                except Exception:
                    raise RuntimeError(
                        "azcli backend is not available (ensure 'az' is installed)"
                    )
            return self._azcli.get_initiative(name)
        else:
            raise ValueError(f"unknown backend: {self.backend}")

    def deploy_initiative(
        self, name: str, initiative: Dict[str, Any]
    ) -> Dict[str, Any]:
        if self.backend == "file":
            return self._file_deploy_initiative(name, initiative)
        elif self.backend == "sdk":
            return self._sdk_deploy_initiative(name, initiative)
        elif self.backend == "azcli":
            if self._azcli is None:
                try:
                    from .azcli_backend import AzCliBackend  # type: ignore

                    self._azcli = AzCliBackend(subscription_id=self.subscription_id)
                except Exception:
                    raise RuntimeError(
                        "azcli backend is not available (ensure 'az' is installed)"
                    )
            return self._azcli.create_or_update_initiative(name, initiative)
        else:
            raise ValueError(f"unknown backend: {self.backend}")

    # -----------------
    # Scope validation and dry-run helpers
    # -----------------
    def _validate_scope(self, scope: str) -> str:
        """Quick heuristic validation for allowed Azure policy assignment scopes.

        Accepts:
        - subscription: /subscriptions/{subscriptionId}
        - resource group: /subscriptions/{subscriptionId}/resourceGroups/{rg}
        - management group: /providers/Microsoft.Management/managementGroups/{mg}
        - tenant or other well-formed Azure resource id starting with a '/'

        Returns the normalized scope (trimmed) or raises ValueError on clearly invalid values.
        """
        if not scope or not isinstance(scope, str):
            raise ValueError("scope must be a non-empty string")
        s = scope.strip()
        # must start with '/'
        if not s.startswith("/"):
            raise ValueError(
                "scope must be a resource id starting with '/' (e.g. /subscriptions/..., /providers/Microsoft.Management/managementGroups/...)"
            )

        # simple patterns
        lower = s.lower()
        if lower.startswith("/subscriptions/"):
            # either subscription root or resource group
            # valid forms: /subscriptions/{id} or /subscriptions/{id}/resourceGroups/{rg}
            parts = s.split("/")
            if len(parts) >= 3 and parts[2]:
                return s
            raise ValueError("invalid subscription scope")
        if lower.startswith("/providers/microsoft.management/managementgroups/"):
            parts = s.split("/")
            # expect at least 5 parts: '', 'providers','Microsoft.Management','managementGroups','{id}'
            if len(parts) >= 5 and parts[4]:
                return s
            raise ValueError("invalid management group scope")

        # fallback: accept any starting with '/'
        return s

    def prepare_assignment_payload(
        self, name: str, assignment: Dict[str, Any], scope: Optional[str] = None
    ) -> Dict[str, Any]:
        """Return the final payload and effective scope that would be used to create/update an assignment.

        Useful for dry-run: returns a dict with keys: scope, name, payload (dict or SDK model converted to dict when possible).
        """
        if scope:
            scope = self._validate_scope(scope)
        # build payload: prefer SDK model when available
        payload = assignment.copy()
        # try to construct SDK model for better representation (but if not available, keep dict)
        try:
            model = self._make_policy_assignment_model(payload)
            as_dict = getattr(model, "as_dict", None)
            if as_dict:
                payload_repr = as_dict()
            elif isinstance(model, dict):
                payload_repr = model
            else:
                payload_repr = payload
        except Exception:
            payload_repr = payload
        return {"scope": scope, "name": name, "payload": payload_repr}

    def _determine_scope_path(self, scope: Optional[str]) -> str:
        """Return the normalized scope path used for ARM REST URLs.

        If scope is provided, return it trimmed; otherwise return subscription-level path if available or empty string.
        """
        if scope:
            return scope.rstrip("/")
        if self.subscription_id:
            return f"/subscriptions/{self.subscription_id}"
        return ""

    # -----------------
    # Assignment public operations
    # -----------------
    def create_assignment(
        self, name: str, assignment: Dict[str, Any], scope: Optional[str] = None
    ) -> Dict[str, Any]:
        """Create an assignment for a policy definition. In file backend it writes into assignments/; in sdk backend it creates the assignment resource."""
        if scope:
            scope = self._validate_scope(scope)
        if self.backend == "file":
            return self._file_deploy_assignment(name, assignment)
        elif self.backend == "sdk":
            return self._sdk_create_assignment(name, assignment, scope)
        elif self.backend == "azcli":
            if self._azcli is None:
                try:
                    from .azcli_backend import AzCliBackend  # type: ignore

                    self._azcli = AzCliBackend(subscription_id=self.subscription_id)
                except Exception:
                    raise RuntimeError(
                        "azcli backend is not available (ensure 'az' is installed)"
                    )
            return self._azcli.create_or_update_assignment(name, assignment, scope)
        else:
            raise ValueError(f"unknown backend: {self.backend}")

    def update_assignment(
        self, name: str, assignment: Dict[str, Any], scope: Optional[str] = None
    ) -> Dict[str, Any]:
        """Update an existing assignment (or create if it doesn't exist)."""
        if scope:
            scope = self._validate_scope(scope)
        # update is equivalent to create in our model
        return self.create_assignment(name, assignment, scope)

    def delete_assignment(
        self, name: str, scope: Optional[str] = None
    ) -> Dict[str, Any]:
        """Delete an assignment by name (and scope)."""
        if scope:
            scope = self._validate_scope(scope)
        if self.backend == "file":
            path = os.path.join(self.assignments_dir, f"{name}{CJSON_EXT}")
            if os.path.exists(path):
                os.remove(path)
                return {"file": path, "status": "deleted"}
            return {"file": path, "status": "not_found"}
        elif self.backend == "sdk":
            return self._sdk_delete_assignment(name, scope)
        elif self.backend == "azcli":
            if self._azcli is None:
                try:
                    from .azcli_backend import AzCliBackend  # type: ignore

                    self._azcli = AzCliBackend(subscription_id=self.subscription_id)
                except Exception:
                    raise RuntimeError(
                        "azcli backend is not available (ensure 'az' is installed)"
                    )
            return self._azcli.delete_assignment(name, scope)
        else:
            raise ValueError(f"unknown backend: {self.backend}")

    def assign_initiative(
        self,
        _initiative_name: str,
        assignment_name: str,
        assignment: Dict[str, Any],
        scope: Optional[str] = None,
    ) -> Dict[str, Any]:
        """Create an assignment for an initiative (policy set). In file backend it writes into assignments/; in sdk backend it creates the assignment resource."""
        if scope:
            scope = self._validate_scope(scope)
        # assignment should reference initiative id in properties; we just create assignment resource
        return self.create_assignment(assignment_name, assignment, scope)

    # -----------------
    # SDK helpers (use azure-mgmt if present, otherwise ARM REST fallbacks)
    # -----------------
    def _make_policy_assignment_model(self, assignment: Dict[str, Any]):
        """Try to build an SDK PolicyAssignment model. Fall back to the dict if not available."""
        # dynamic import to avoid static analyzer warnings
        sdk_policy_assignment = _import_attr(
            _AZURE_MGMT_POLICY_MODELS, "PolicyAssignment"
        )
        if sdk_policy_assignment is None:
            # fallback to alternate location
            policy_mod = _import_attr(_AZURE_MGMT_RESOURCE, "policy")
            if policy_mod:
                sdk_policy_assignment = getattr(
                    getattr(policy_mod, "models", {}), "PolicyAssignment", None
                )

        if sdk_policy_assignment:
            try:
                return sdk_policy_assignment(**assignment)
            except Exception:
                # model construction failed; return raw dict
                return assignment
        return assignment

    def _sdk_create_assignment(
        self, name: str, assignment: Dict[str, Any], scope: Optional[str] = None
    ) -> Dict[str, Any]:
        """Create an assignment using the SDK (azure-mgmt) if available, otherwise fallback to ARM REST API."""
        policy_client_cls = _import_attr(_AZURE_MGMT_RESOURCE, "PolicyClient")
        if policy_client_cls:
            client = policy_client_cls(self.credential, self.subscription_id)
            params = assignment.copy()
            model = self._make_policy_assignment_model(params)
            res = client.policy_assignments.create(scope, name, model)
            as_dict = getattr(res, "as_dict", None)
            return as_dict() if as_dict else getattr(res, "__dict__", {})
        # ARM REST fallback
        if not self.credential:
            raise RuntimeError(_CREDENTIAL_ERR)
        api = "2021-06-01"
        scope_path = self._determine_scope_path(scope)
        url = f"https://management.azure.com{scope_path}/providers/Microsoft.Authorization/policyAssignments/{name}?api-version={api}"
        return self._arm_request("PUT", url, json=assignment.copy())

    def _sdk_delete_assignment(
        self, name: str, scope: Optional[str] = None
    ) -> Dict[str, Any]:
        """Delete an assignment using the SDK (azure-mgmt) if available, otherwise fallback to ARM REST API."""
        policy_client_cls = _import_attr(_AZURE_MGMT_RESOURCE, "PolicyClient")
        if policy_client_cls:
            client = policy_client_cls(self.credential, self.subscription_id)
            res = client.policy_assignments.delete(scope, name)
            return {"status": "deleted", "raw": getattr(res, "as_dict", lambda: {})()}
        if not self.credential:
            raise RuntimeError(_CREDENTIAL_ERR)
        api = "2021-06-01"
        scope_path = self._determine_scope_path(scope)
        url = f"https://management.azure.com{scope_path}/providers/Microsoft.Authorization/policyAssignments/{name}?api-version={api}"
        try:
            resp = self._arm_request("DELETE", url)
            return {"status": "deleted", "response": resp}
        except Exception as ex:
            return {"status": "error", "error": str(ex)}

    def _sdk_list_assignments(
        self, scope: Optional[str] = None
    ) -> List[Dict[str, Any]]:
        """List assignments using the SDK (azure-mgmt) if available, otherwise fallback to ARM REST API."""
        policy_client_cls = _import_attr(_AZURE_MGMT_RESOURCE, "PolicyClient")
        if policy_client_cls:
            client = policy_client_cls(self.credential, self.subscription_id)
            if scope:
                items = list(client.policy_assignments.list_for_scope(scope))
            else:
                try:
                    items = list(client.policy_assignments.list())
                except Exception:
                    items = list(
                        client.policy_assignments.list_for_subscription(
                            self.subscription_id
                        )
                    )
            out: List[Dict[str, Any]] = []
            for it in items:
                as_dict = getattr(it, "as_dict", None)
                out.append(as_dict() if as_dict else getattr(it, "__dict__", {}))
            return out
        # ARM REST fallback
        if not self.credential:
            raise RuntimeError(_CREDENTIAL_ERR)
        api = "2021-06-01"
        scope_path = self._determine_scope_path(scope)
        url = f"https://management.azure.com{scope_path}/providers/Microsoft.Authorization/policyAssignments?api-version={api}"
        resp = self._arm_request("GET", url)
        return resp.get("value", [])

    # -----------------
    # Minimal policy and initiative SDK helpers (used by other features/tests)
    # -----------------
    def _get_arm_token(self) -> str:
        """Acquire an Azure AD token for ARM (management) if a credential is available."""
        if not self.credential:
            raise RuntimeError(
                "Azure credential not provided. Pass an azure.identity credential to AzurePolicyEngine(credential=...) when using the sdk backend."
            )
        token = self.credential.get_token("https://management.azure.com/.default")
        return token.token

    def _arm_request(self, method: str, url: str, **kwargs) -> Dict[str, Any]:
        """Helper for ARM REST calls using requests and an acquired bearer token."""
        try:
            import requests
        except Exception:
            raise RuntimeError(
                "requests library is required for sdk backend; add 'requests' to your environment"
            )
        token = self._get_arm_token()
        headers = kwargs.pop("headers", {})
        headers.update(
            {"Authorization": f"Bearer {token}", "Content-Type": "application/json"}
        )
        r = requests.request(method, url, headers=headers, **kwargs)
        try:
            body = r.json()
        except Exception:
            body = {"raw_text": r.text}
        if not r.ok:
            raise RuntimeError(
                f"ARM request {method} {url} failed: {r.status_code} {r.reason} - {body}"
            )
        return body

    def _sdk_list_policies(self) -> List[Dict[str, Any]]:
        """List policy definitions using the SDK (azure-mgmt) if available, otherwise fallback to ARM REST API."""
        policy_client_cls = _import_attr(_AZURE_MGMT_RESOURCE, "PolicyClient")
        if policy_client_cls:
            client = policy_client_cls(self.credential, self.subscription_id)
            items = list(client.policy_definitions.list())
            out: List[Dict[str, Any]] = []
            for it in items:
                as_dict = getattr(it, "as_dict", None)
                out.append(as_dict() if as_dict else getattr(it, "__dict__", {}))
            return out
        if not self.credential:
            raise RuntimeError("Azure credential not provided for SDK/ARM calls")
        api = "2021-06-01"
        scope = f"/subscriptions/{self.subscription_id}" if self.subscription_id else ""
        url = f"https://management.azure.com{scope}/providers/Microsoft.Authorization/policyDefinitions?api-version={api}"
        resp = self._arm_request("GET", url)
        return resp.get("value", [])

    def _sdk_get_policy(self, name: str) -> Dict[str, Any]:
        """Get a single policy definition by name using the SDK (azure-mgmt) if available, otherwise fallback to ARM REST API."""
        policy_client_cls = _import_attr(_AZURE_MGMT_RESOURCE, "PolicyClient")
        if policy_client_cls:
            client = policy_client_cls(self.credential, self.subscription_id)
            item = client.policy_definitions.get(name)
            as_dict = getattr(item, "as_dict", None)
            return as_dict() if as_dict else getattr(item, "__dict__", {})
        api = "2021-06-01"
        scope = f"/subscriptions/{self.subscription_id}" if self.subscription_id else ""
        url = f"https://management.azure.com{scope}/providers/Microsoft.Authorization/policyDefinitions/{name}?api-version={api}"
        return self._arm_request("GET", url)

    def _sdk_deploy_policy(self, name: str, policy: Dict[str, Any]) -> Dict[str, Any]:
        """Create or update a policy definition using the SDK (azure-mgmt) if available, otherwise fallback to ARM REST API."""
        policy_client_cls = _import_attr(_AZURE_MGMT_RESOURCE, "PolicyClient")
        if policy_client_cls:
            client = policy_client_cls(self.credential, self.subscription_id)
            body = policy.copy()
            body.pop("name", None)
            model = self._make_policy_definition_model(body)
            try:
                res = client.policy_definitions.create_or_update(name, model)
            except Exception:
                res = client.policy_definitions.create(name, model)
            as_dict = getattr(res, "as_dict", None)
            return as_dict() if as_dict else getattr(res, "__dict__", {})
        api = "2021-06-01"
        scope = f"/subscriptions/{self.subscription_id}" if self.subscription_id else ""
        url = f"https://management.azure.com{scope}/providers/Microsoft.Authorization/policyDefinitions/{name}?api-version={api}"
        body = policy.copy()
        body.pop("name", None)
        return self._arm_request("PUT", url, json=body)

    def _make_policy_definition_model(self, policy: Dict[str, Any]):
        """Construct a policy definition model object for SDK use, or return the original policy dict."""
        sdk_def = _import_attr(_AZURE_MGMT_POLICY_MODELS, "PolicyDefinition")
        if sdk_def is None:
            policy_mod = _import_attr(_AZURE_MGMT_RESOURCE, "policy")
            if policy_mod:
                sdk_def = getattr(
                    getattr(policy_mod, "models", {}), "PolicyDefinition", None
                )
        if sdk_def:
            try:
                return sdk_def(**policy)
            except Exception:
                return policy
        return policy

    def _sdk_list_initiatives(self) -> List[Dict[str, Any]]:
        """List initiative (policy set) definitions using the SDK (azure-mgmt) if available, otherwise fallback to ARM REST API."""
        policy_client_cls = _import_attr(_AZURE_MGMT_RESOURCE, "PolicyClient")
        if policy_client_cls:
            client = policy_client_cls(self.credential, self.subscription_id)
            items = list(client.policy_set_definitions.list())
            out: List[Dict[str, Any]] = []
            for it in items:
                as_dict = getattr(it, "as_dict", None)
                out.append(as_dict() if as_dict else getattr(it, "__dict__", {}))
            return out
        api = "2021-06-01"
        scope = f"/subscriptions/{self.subscription_id}" if self.subscription_id else ""
        url = f"https://management.azure.com{scope}/providers/Microsoft.Authorization/policySetDefinitions?api-version={api}"
        resp = self._arm_request("GET", url)
        return resp.get("value", [])

    def _sdk_get_initiative(self, name: str) -> Dict[str, Any]:
        """Get a single initiative (policy set) definition by name using the SDK (azure-mgmt) if available, otherwise fallback to ARM REST API."""
        policy_client_cls = _import_attr(_AZURE_MGMT_RESOURCE, "PolicyClient")
        if policy_client_cls:
            client = policy_client_cls(self.credential, self.subscription_id)
            if name.startswith("/"):
                # azure-mgmt client typically needs name, so fallback to ARM
                raise RuntimeError("full resource-id get not supported by SDK path")
            item = client.policy_set_definitions.get(name)
            as_dict = getattr(item, "as_dict", None)
            return as_dict() if as_dict else getattr(item, "__dict__", {})
        api = "2021-06-01"
        scope = f"/subscriptions/{self.subscription_id}" if self.subscription_id else ""
        if name.startswith("/"):
            url = f"https://management.azure.com{name}?api-version={api}"
        else:
            url = f"https://management.azure.com{scope}/providers/Microsoft.Authorization/policySetDefinitions/{name}?api-version={api}"
        return self._arm_request("GET", url)

    def _sdk_deploy_initiative(
        self, name: str, initiative: Dict[str, Any]
    ) -> Dict[str, Any]:
        """Create or update an initiative (policy set) definition using the SDK (azure-mgmt) if available, otherwise fallback to ARM REST API."""
        policy_client_cls = _import_attr(_AZURE_MGMT_RESOURCE, "PolicyClient")
        if policy_client_cls:
            client = policy_client_cls(self.credential, self.subscription_id)
            body = initiative.copy()
            body.pop("name", None)
            model = self._make_policy_set_definition_model(body)
            try:
                res = client.policy_set_definitions.create_or_update(name, model)
            except Exception:
                res = client.policy_set_definitions.create(name, model)
            as_dict = getattr(res, "as_dict", None)
            return as_dict() if as_dict else getattr(res, "__dict__", {})
        api = "2021-06-01"
        scope = f"/subscriptions/{self.subscription_id}" if self.subscription_id else ""
        url = f"https://management.azure.com{scope}/providers/Microsoft.Authorization/policySetDefinitions/{name}?api-version={api}"
        body = initiative.copy()
        body.pop("name", None)
        return self._arm_request("PUT", url, json=body)

    def _make_policy_set_definition_model(self, initiative: Dict[str, Any]):
        """Construct a policy set definition model object for SDK use, or return the original initiative dict."""
        sdk_cls = _import_attr(_AZURE_MGMT_POLICY_MODELS, "PolicySetDefinition")
        if sdk_cls is None:
            policy_mod = _import_attr(_AZURE_MGMT_RESOURCE, "policy")
            if policy_mod:
                sdk_cls = getattr(
                    getattr(policy_mod, "models", {}), "PolicySetDefinition", None
                )
        if sdk_cls:
            try:
                return sdk_cls(**initiative)
            except Exception:
                return initiative
        return initiative


# end of engine.py



from __future__ import annotations

import json
import os
import subprocess
from types import SimpleNamespace
from unittest.mock import patch

import pytest

from py_azpe.azcli_backend import AzCliBackend, AzCliError


def _make_completed_process(stdout: str = "{}", stderr: str = ""):
    return SimpleNamespace(stdout=stdout, stderr=stderr)


@patch("py_azpe.azcli_backend.subprocess.run")
def test_list_policy_definitions_subscription(mock_run):
    # simulate az rest returning a list with one policy definition
    resp = {"value": [{"name": "p1", "properties": {"displayName": "P1"}}]}
    mock_run.return_value = _make_completed_process(json.dumps(resp), "")

    b = AzCliBackend(subscription_id="0000")
    res = b.list_policy_definitions()

    assert isinstance(res, list)
    assert len(res) == 1
    assert res[0]["name"] == "p1"
    # assert subprocess.run was called with az and rest and subscription id in URI
    called_cmd = mock_run.call_args[0][0]
    assert called_cmd[0] == "az"
    assert "rest" in called_cmd
    # find the uri arg
    uri_items = [
        c for c in called_cmd if isinstance(c, str) and "management.azure.com" in c
    ]
    assert uri_items, "no management.azure.com URI found in az command"
    assert "/subscriptions/0000/" in uri_items[0]
    assert "policyDefinitions" in uri_items[0]


@patch("py_azpe.azcli_backend.subprocess.run")
def test_create_policy_definition_writes_body_and_parses_response(mock_run, tmp_path):
    # simulate az rest returning the created policy definition
    resp = {"name": "p2", "properties": {"displayName": "P2"}}
    mock_run.return_value = _make_completed_process(json.dumps(resp), "")

    b = AzCliBackend(subscription_id="0000")
    body = {"name": "p2", "properties": {"displayName": "P2"}}
    res = b.create_or_update_policy_definition("p2", body)

    assert isinstance(res, dict)
    assert res.get("name") == "p2"

    # ensure subprocess was called; inspect args to find temp file created with @ prefix
    called_cmd = mock_run.call_args[0][0]
    # the --body value will be passed as a separate element like '@/tmp/tmpXXXXX.json'
    body_items = [c for c in called_cmd if isinstance(c, str) and c.startswith("@")]
    assert body_items, "expected a @tmpfile argument in az call"
    tmpfile = body_items[0][1:]
    # temporary file should have been removed by the backend
    assert not os.path.exists(tmpfile)


@patch("py_azpe.azcli_backend.subprocess.run")
def test_create_and_delete_assignment(mock_run):
    # We'll simulate two separate subprocess calls: one for PUT (create) and one for DELETE
    create_resp = {"name": "assign1", "properties": {"displayName": "Assign 1"}}
    delete_resp = {}

    # configure side_effect to return different results per call
    mock_run.side_effect = [
        _make_completed_process(json.dumps(create_resp), ""),
        _make_completed_process(json.dumps(delete_resp), ""),
    ]

    b = AzCliBackend(subscription_id="0000")
    assignment = {
        "name": "assign1",
        "properties": {"displayName": "Assign 1", "scope": "/subscriptions/0000"},
    }

    # create
    res_create = b.create_or_update_assignment(
        "assign1", assignment, scope="/subscriptions/0000"
    )
    assert isinstance(res_create, dict)
    assert res_create.get("name") == "assign1"

    # inspect the first call (PUT)
    first_cmd = mock_run.call_args_list[0][0][0]
    uri_items = [
        c for c in first_cmd if isinstance(c, str) and "management.azure.com" in c
    ]
    assert uri_items and "policyAssignments" in uri_items[0]
    assert "/subscriptions/0000" in uri_items[0]
    # find temp body file
    body_items = [c for c in first_cmd if isinstance(c, str) and c.startswith("@")]
    assert body_items
    tmpfile = body_items[0][1:]
    # ensure temp file has been removed by backend
    assert not os.path.exists(tmpfile)

    # delete
    res_del = b.delete_assignment("assign1", scope="/subscriptions/0000")
    assert res_del == {}

    # inspect second call (DELETE)
    second_cmd = mock_run.call_args_list[1][0][0]
    uri_items2 = [
        c for c in second_cmd if isinstance(c, str) and "management.azure.com" in c
    ]
    assert uri_items2 and "policyAssignments" in uri_items2[0]
    assert "/subscriptions/0000" in uri_items2[0]


# New tests: error handling and tenant-scoped listing
@patch("py_azpe.azcli_backend.subprocess.run")
def test_azcli_raises_on_calledprocesserror(mock_run):
    # Simulate az failing with a CalledProcessError
    mock_run.side_effect = subprocess.CalledProcessError(
        returncode=2, cmd=["az", "rest"], stderr="boom error"
    )

    b = AzCliBackend(subscription_id="0000")
    with pytest.raises(AzCliError) as exc:
        b.list_policy_definitions()
    assert "boom error" in str(exc.value)


@patch("py_azpe.azcli_backend.subprocess.run")
def test_non_json_stdout_returns_raw(mock_run):
    # Simulate az returning non-JSON stdout
    mock_run.return_value = _make_completed_process("OK", "")

    b = AzCliBackend(subscription_id="0000")
    # call internal runner directly to observe behavior
    res = b._run_az(
        [
            "rest",
            "--method",
            "GET",
            "--uri",
            "https://management.azure.com/providers/Microsoft.Authorization/policyDefinitions?api-version=2021-06-01",
        ]
    )
    assert res == {"raw": "OK"}


@patch("py_azpe.azcli_backend.subprocess.run")
def test_list_policy_definitions_tenant_scope(mock_run):
    # simulate tenant-scoped list when no subscription id is provided
    resp = {"value": [{"name": "t1", "properties": {"displayName": "T1"}}]}
    mock_run.return_value = _make_completed_process(json.dumps(resp), "")

    b = AzCliBackend(subscription_id=None)
    res = b.list_policy_definitions()

    assert isinstance(res, list) and len(res) == 1
    assert res[0]["name"] == "t1"
    called_cmd = mock_run.call_args[0][0]
    uri_items = [
        c for c in called_cmd if isinstance(c, str) and "management.azure.com" in c
    ]
    assert uri_items
    # tenant-scoped should not include /subscriptions/
    assert "/subscriptions/" not in uri_items[0]
    assert "policyDefinitions" in uri_items[0]


if __name__ == "__main__":
    pytest.main([__file__])



import os
from unittest import mock

import importlib


def test_service_principal_credential(tmp_path, monkeypatch):
    # ensure env vars are set for service principal
    monkeypatch.setenv("AZURE_CLIENT_ID", "cid")
    monkeypatch.setenv("AZURE_CLIENT_SECRET", "csecret")
    monkeypatch.setenv("AZURE_TENANT_ID", "ctenant")
    monkeypatch.setenv("AZURE_SUBSCRIPTION_ID", "subid")

    # create a fake azure.identity module with ClientSecretCredential
    fake_module = mock.MagicMock()

    class FakeCred:
        def __init__(self, tenant_id=None, client_id=None, client_secret=None):
            self.tenant_id = tenant_id
            self.client_id = client_id
            self.client_secret = client_secret

    fake_module.ClientSecretCredential = FakeCred

    with mock.patch.dict("sys.modules", {"azure.identity": fake_module}):
        # reload the module under test to pick up the patched azure.identity
        mod = importlib.import_module("py_azpe.azure_auth")
        importlib.reload(mod)
        cred = mod.get_credential()
        assert isinstance(cred, FakeCred)
        assert cred.client_id == "cid"
        assert cred.client_secret == "csecret"
        assert cred.tenant_id == "ctenant"


def test_default_credential_fallback(monkeypatch):
    # ensure service principal env vars are not set
    monkeypatch.delenv("AZURE_CLIENT_ID", raising=False)
    monkeypatch.delenv("AZURE_CLIENT_SECRET", raising=False)
    monkeypatch.delenv("AZURE_TENANT_ID", raising=False)
    monkeypatch.setenv("AZURE_SUBSCRIPTION_ID", "subid")

    fake_module = mock.MagicMock()

    class FakeDefault:
        def __init__(self):
            pass

    fake_module.DefaultAzureCredential = FakeDefault

    with mock.patch.dict("sys.modules", {"azure.identity": fake_module}):
        mod = importlib.import_module("py_azpe.azure_auth")
        importlib.reload(mod)
        cred = mod.get_credential()
        assert isinstance(cred, FakeDefault)


# end of test_azure_auth.py



# conftest.py intentionally left minimal for CI hygiene.
# The project previously added sys.path modifications here to make pytest import the package
# during local development. For CI we prefer to install the package in editable mode
# (pip install -e .) so tests run with the package properly installed instead of
# mutating sys.path.
#
# If you run tests locally and face import errors, run:
#   python -m venv .venv
#   source .venv/bin/activate
#   pip install -e .
#   pip install -r requirements.txt
# and then run `pytest` from the project root.



"""
Utility to read Azure environment variables and provide authenticated Azure SDK clients.

Usage:
- Set AZURE_CLIENT_ID, AZURE_CLIENT_SECRET, AZURE_TENANT_ID for a service principal.
- Or rely on DefaultAzureCredential (Managed Identity, Azure CLI, etc.).
- AZURE_SUBSCRIPTION_ID is required for subscription-scoped clients.

The module avoids hard dependency on azure packages at import time; it performs lazy imports and provides informative errors
if the package isn't installed.
"""
from __future__ import annotations

import os
from typing import Optional


def _get_env(key: str, required: bool = True) -> Optional[str]:
    val = os.getenv(key)
    if required and not val:
        raise RuntimeError(f"Missing required environment variable: {key}")
    return val


def get_subscription_id() -> str:
    """Return AZURE_SUBSCRIPTION_ID or raise RuntimeError if unset."""
    return _get_env("AZURE_SUBSCRIPTION_ID")


def get_credential():
    """
    Return an Azure credential instance.

    Priority:
    - If AZURE_CLIENT_ID, AZURE_CLIENT_SECRET and AZURE_TENANT_ID are set, return ClientSecretCredential.
    - Otherwise return DefaultAzureCredential.

    Raises RuntimeError if required packages are missing.
    """
    client_id = os.getenv("AZURE_CLIENT_ID")
    client_secret = os.getenv("AZURE_CLIENT_SECRET")
    tenant_id = os.getenv("AZURE_TENANT_ID")

    if client_id and client_secret and tenant_id:
        try:
            from azure.identity import ClientSecretCredential  # type: ignore[import]
        except Exception as exc:  # ImportError or similar
            raise RuntimeError(
                "azure-identity is required to use ClientSecretCredential. Install with 'pip install azure-identity'."
            ) from exc

        return ClientSecretCredential(
            tenant_id=tenant_id, client_id=client_id, client_secret=client_secret
        )

    # fallback
    try:
        from azure.identity import DefaultAzureCredential  # type: ignore[import]
    except Exception as exc:
        raise RuntimeError(
            "azure-identity is required for DefaultAzureCredential. Install with 'pip install azure-identity'."
        ) from exc

    return DefaultAzureCredential()


def create_resource_client(credential=None, subscription_id: Optional[str] = None):
    """
    Create and return an instance of azure.mgmt.resource.ResourceManagementClient.

    If credential or subscription_id are not provided, they will be read using helpers above.
    """
    if credential is None:
        credential = get_credential()
    if subscription_id is None:
        subscription_id = get_subscription_id()

    try:
        from azure.mgmt.resource import ResourceManagementClient  # type: ignore[import]
    except Exception as exc:
        raise RuntimeError(
            "azure-mgmt-resource is required to create ResourceManagementClient. Install with 'pip install azure-mgmt-resource'."
        ) from exc

    return ResourceManagementClient(credential, subscription_id)


def create_policyinsights_client(
    credential=None, subscription_id: Optional[str] = None
):
    """
    Create and return an instance of azure.mgmt.policyinsights.PolicyInsightsClient.

    Some policy clients are not subscription-scoped and may not require subscription_id but accepting it keeps API parity.
    """
    if credential is None:
        credential = get_credential()

    try:
        from azure.mgmt.policyinsights import PolicyInsightsClient  # type: ignore[import]
    except Exception as exc:
        raise RuntimeError(
            "azure-mgmt-policyinsights is required to create PolicyInsightsClient. Install with 'pip install azure-mgmt-policyinsights'."
        ) from exc

    # PolicyInsightsClient doesn't require subscription in constructor
    return PolicyInsightsClient(credential)


if __name__ == "__main__":
    # quick smoke test
    try:
        cred = get_credential()
        sub = get_subscription_id()
        print(f"Using subscription: {sub}")
        # instantiate clients to ensure imports succeed
        _ = create_resource_client(credential=cred, subscription_id=sub)
        _ = create_policyinsights_client(credential=cred)
        print("Clients created successfully.")
    except Exception as e:
        print("Error during smoke test:", e)



import importlib
from unittest import mock


def test_create_resource_client_uses_azure_mgmt_resource(monkeypatch):
    # prepare fake ResourceManagementClient class
    fake_module = mock.MagicMock()

    class FakeRMC:
        def __init__(self, credential, subscription_id):
            self.credential = credential
            self.subscription_id = subscription_id

    fake_module.ResourceManagementClient = FakeRMC

    with mock.patch.dict("sys.modules", {"azure.mgmt.resource": fake_module}):
        mod = importlib.import_module("py_azpe.azure_auth")
        importlib.reload(mod)
        # provide a fake credential object
        fake_cred = object()
        client = mod.create_resource_client(
            credential=fake_cred, subscription_id="subid"
        )
        assert isinstance(client, FakeRMC)
        assert client.credential is fake_cred
        assert client.subscription_id == "subid"


def test_create_policyinsights_client_uses_azure_mgmt_policyinsights(monkeypatch):
    fake_module = mock.MagicMock()

    class FakePIC:
        def __init__(self, credential):
            self.credential = credential

    fake_module.PolicyInsightsClient = FakePIC

    with mock.patch.dict("sys.modules", {"azure.mgmt.policyinsights": fake_module}):
        mod = importlib.import_module("py_azpe.azure_auth")
        importlib.reload(mod)
        fake_cred = object()
        client = mod.create_policyinsights_client(credential=fake_cred)
        assert isinstance(client, FakePIC)
        assert client.credential is fake_cred



import os
import pytest

from py_azpe import AzurePolicyEngine


@pytest.mark.skipif(
    os.environ.get("AZPE_RUN_AZURE_INTEGRATION", "false").lower() != "true",
    reason="Integration tests disabled by default",
)
def test_deploy_policy_and_assignment_live():
    # This test runs only when AZPE_RUN_AZURE_INTEGRATION=true and valid credentials are available.
    subscription = os.environ.get("AZURE_SUBSCRIPTION_ID")
    assert subscription, "AZURE_SUBSCRIPTION_ID required for integration test"

    cred = None
    try:
        from azure.identity import DefaultAzureCredential

        cred = DefaultAzureCredential()
    except Exception:
        pytest.skip("azure-identity not available or configured")

    engine = AzurePolicyEngine(
        backend="sdk", subscription_id=subscription, credential=cred
    )

    # create a policy definition (name is arbitrary)
    pol = {
        "properties": {
            "displayName": "IT test policy",
            "policyRule": {
                "if": {"field": "location", "equals": ""},
                "then": {"effect": "audit"},
            },
        }
    }
    res = engine.deploy_policy("it_test_policy", pol)
    assert isinstance(res, dict)

    # prepare an assignment payload as a dry-run
    assign = {
        "properties": {
            "displayName": "IT test assign",
            "policyDefinitionId": res.get("id") or res.get("name"),
        }
    }
    payload = engine.prepare_assignment_payload(
        "it_test_assign", assign, scope=f"/subscriptions/{subscription}"
    )
    assert payload["name"] == "it_test_assign"



from __future__ import annotations

import json
import os
from typing import Dict

from .engine import AzurePolicyEngine


def _read_body(path: str) -> Dict:
    with open(path, "r", encoding="utf-8") as f:
        return json.load(f)


def _print(obj: object) -> None:
    print(json.dumps(obj, indent=2))


# Handler implementations (small, focused)
def _handle_list(args, engine: AzurePolicyEngine) -> None:
    _print(engine.list_policies())


def _handle_get(args, engine: AzurePolicyEngine) -> None:
    _print(engine.get_policy(args.name))


def _handle_deploy(args, engine: AzurePolicyEngine) -> None:
    body = _read_body(args.file)
    _print(engine.deploy_policy(args.name, body))


def _handle_list_assignments(args, engine: AzurePolicyEngine) -> None:
    _print(engine.list_assignments())


def _handle_create_assignment(args, engine: AzurePolicyEngine) -> None:
    common_scope = args.scope or (
        f"/providers/Microsoft.Management/managementGroups/{getattr(args, 'mgmt_group', None)}"
        if getattr(args, "mgmt_group", None)
        else None
    )
    body = _read_body(args.file)
    if getattr(args, "dry_run", False):
        _print(
            {
                "dry_run": True,
                "action": "create_assignment",
                **engine.prepare_assignment_payload(
                    args.name, body, scope=common_scope
                ),
            }
        )
    else:
        _print(engine.create_assignment(args.name, body, scope=common_scope))


def _handle_update_assignment(args, engine: AzurePolicyEngine) -> None:
    common_scope = args.scope or (
        f"/providers/Microsoft.Management/managementGroups/{getattr(args, 'mgmt_group', None)}"
        if getattr(args, "mgmt_group", None)
        else None
    )
    body = _read_body(args.file)
    if getattr(args, "dry_run", False):
        _print(
            {
                "dry_run": True,
                "action": "update_assignment",
                **engine.prepare_assignment_payload(
                    args.name, body, scope=common_scope
                ),
            }
        )
    else:
        _print(engine.update_assignment(args.name, body, scope=common_scope))


def _handle_delete_assignment(args, engine: AzurePolicyEngine) -> None:
    common_scope = args.scope or (
        f"/providers/Microsoft.Management/managementGroups/{getattr(args, 'mgmt_group', None)}"
        if getattr(args, "mgmt_group", None)
        else None
    )
    if getattr(args, "dry_run", False):
        _print(
            {
                "dry_run": True,
                "action": "delete_assignment",
                "scope": common_scope,
                "name": args.name,
            }
        )
    else:
        _print(engine.delete_assignment(args.name, scope=common_scope))


def _handle_add_assignment(args, engine: AzurePolicyEngine) -> None:
    if args.backend != "file":
        import sys

        print("add-assignment only supported for file backend", file=sys.stderr)
        sys.exit(2)
    body = _read_body(args.file)
    dest = os.path.join(engine.assignments_dir, f"{args.name}.json")
    with open(dest, "w", encoding="utf-8") as f:
        json.dump(body, f, indent=2)
    _print({"file": dest, "status": "written"})


def _handle_list_initiatives(args, engine: AzurePolicyEngine) -> None:
    res = engine.list_initiatives()
    if getattr(args, "raw", False):
        _print(res)
    else:
        print("Initiatives:")
        for it in res:
            print(" - ", it.get("name"))


def _handle_assign_initiative(args, engine: AzurePolicyEngine) -> None:
    common_scope = args.scope or (
        f"/providers/Microsoft.Management/managementGroups/{getattr(args, 'mgmt_group', None)}"
        if getattr(args, "mgmt_group", None)
        else None
    )
    body = _read_body(args.file)
    if getattr(args, "dry_run", False):
        _print(
            {
                "dry_run": True,
                "action": "assign_initiative",
                "initiative": args.initiative_name,
                **engine.prepare_assignment_payload(
                    args.assignment_name, body, scope=common_scope
                ),
            }
        )
    else:
        _print(
            engine.assign_initiative(
                args.initiative_name, args.assignment_name, body, scope=common_scope
            )
        )


# Dispatch map keeps handle_command small and easy to extend
_DISPATCH = {
    "list": _handle_list,
    "get": _handle_get,
    "deploy": _handle_deploy,
    "assignments": _handle_list_assignments,
    "create-assignment": _handle_create_assignment,
    "update-assignment": _handle_update_assignment,
    "delete-assignment": _handle_delete_assignment,
    "add-assignment": _handle_add_assignment,
    "list-initiatives": _handle_list_initiatives,
    "assign-initiative": _handle_assign_initiative,
}


def handle_command(args, engine: AzurePolicyEngine) -> None:
    """Dispatch and perform the requested command. This module keeps CLI main small and testable."""
    cmd = getattr(args, "cmd", None)
    handler = _DISPATCH.get(cmd)
    if handler:
        handler(args, engine)
    else:
        # unknown command -> print help-ish information
        print("unknown command or handler not implemented")



import types

from py_azpe import AzurePolicyEngine


class DummyAssignmentsClient:
    def __init__(self):
        self.created = []

    def create(self, scope, name, params):
        self.created.append((scope, name, params))
        return types.SimpleNamespace(
            as_dict=lambda: {
                "scope": scope,
                "name": name,
                "params_type": type(params).__name__,
            }
        )


class DummyPolicyClient:
    def __init__(self):
        self.policy_assignments = DummyAssignmentsClient()


def _patch_policy_client(monkeypatch, dummy):
    monkeypatch.setitem(
        __import__("sys").modules,
        "azure.mgmt.resource",
        types.SimpleNamespace(PolicyClient=lambda cred, sub: dummy),
    )


def test_create_assignment_subscription_scope(monkeypatch, tmp_path):
    engine = AzurePolicyEngine(backend="sdk", policy_dir=str(tmp_path))
    dummy = DummyPolicyClient()
    _patch_policy_client(monkeypatch, dummy)

    assign = {
        "properties": {
            "displayName": "Sub Assign",
            "policyDefinitionId": "/subscriptions/0000/providers/Microsoft.Authorization/policyDefinitions/pd1",
        }
    }
    scope = "/subscriptions/0000"
    _ = engine.create_assignment("subAssign", assign, scope=scope)

    assert dummy.policy_assignments.created
    cscope, cname, _ = dummy.policy_assignments.created[0]
    assert cscope == scope
    assert cname == "subAssign"


def test_create_assignment_resource_group_scope(monkeypatch, tmp_path):
    engine = AzurePolicyEngine(backend="sdk", policy_dir=str(tmp_path))
    dummy = DummyPolicyClient()
    _patch_policy_client(monkeypatch, dummy)

    assign = {
        "properties": {
            "displayName": "RG Assign",
            "policyDefinitionId": "/subscriptions/0000/resourceGroups/rg1/providers/Microsoft.Authorization/policyDefinitions/pd1",
        }
    }
    scope = "/subscriptions/0000/resourceGroups/rg1"
    _ = engine.create_assignment("rgAssign", assign, scope=scope)

    assert dummy.policy_assignments.created
    cscope, cname, _ = dummy.policy_assignments.created[0]
    assert cscope == scope
    assert cname == "rgAssign"



import os
import json

from py_azpe import AzurePolicyEngine


def test_file_backend_roundtrip(tmp_path):
    base = tmp_path / "pols"
    base.mkdir()
    engine = AzurePolicyEngine(backend="file", policy_dir=str(base))
    sample = {"name": "testpolicy", "properties": {"displayName": "Test"}}
    engine.deploy_policy("testpolicy", sample)
    got = engine.get_policy("testpolicy")
    assert (
        got["name"] == "testpolicy"
        or got.get("properties", {}).get("displayName") == "Test"
    )
    listed = engine.list_policies()
    assert any(
        p.get("name") == "testpolicy"
        or os.path.basename(p.get("file", "")).startswith("testpolicy")
        for p in listed
    )


def test_assignments_and_initiatives(tmp_path):
    base = tmp_path / "pols"
    base.mkdir()
    engine = AzurePolicyEngine(backend="file", policy_dir=str(base))
    assign = {
        "name": "assign1",
        "properties": {"displayName": "Assign 1", "scope": "/subscriptions/0000"},
    }
    assign_path = os.path.join(engine.assignments_dir, "assign1.json")
    with open(assign_path, "w", encoding="utf-8") as f:
        json.dump(assign, f)
    assigns = engine.list_assignments()
    assert any(
        a.get("name") == "assign1"
        or os.path.basename(a.get("file", "")).startswith("assign1")
        for a in assigns
    )

    init = {
        "name": "init1",
        "properties": {"displayName": "Init 1", "policyDefinitions": []},
    }
    init_path = os.path.join(engine.initiatives_dir, "init1.json")
    with open(init_path, "w", encoding="utf-8") as f:
        json.dump(init, f)
    inits = engine.list_initiatives()
    assert any(
        i.get("name") == "init1"
        or os.path.basename(i.get("file", "")).startswith("init1")
        for i in inits
    )



import types
import pytest
import os
import json

from py_azpe import AzurePolicyEngine


class DummyModel:
    def __init__(self, **kwargs):
        self._data = kwargs

    def as_dict(self):
        return self._data


class DummyAssignmentsClient:
    def __init__(self):
        self.created = []

    def create(self, scope, name, params):
        # capture the type of params
        self.created.append((scope, name, params))
        # return an object with as_dict
        return types.SimpleNamespace(
            as_dict=lambda: {
                "scope": scope,
                "name": name,
                "params_type": type(params).__name__,
            }
        )


class DummyPolicyClient:
    def __init__(self):
        self.policy_assignments = DummyAssignmentsClient()


class DummyPolicyClientFactory:
    def __init__(self, client):
        self._client = client

    def __call__(self, credential, subscription):
        return self._client


def test_sdk_create_assignment_uses_model(monkeypatch, tmp_path):
    # Build engine with sdk backend
    engine = AzurePolicyEngine(backend="sdk", policy_dir=str(tmp_path))

    # Prepare dummy client and patch the import path
    dummy_client = DummyPolicyClient()

    # Mock PolicyClient factory in azure.mgmt.resource
    fake_pkg = types.SimpleNamespace(
        PolicyClient=DummyPolicyClientFactory(dummy_client)
    )
    monkeypatch.setitem(__import__("sys").modules, "azure.mgmt.resource", fake_pkg)

    # Also provide a dummy model class at azure.mgmt.resource.policy.models.PolicyAssignment
    policy_module = types.SimpleNamespace()
    policy_module.models = types.SimpleNamespace(PolicyAssignment=DummyModel)
    monkeypatch.setitem(
        __import__("sys").modules, "azure.mgmt.resource.policy", policy_module
    )

    # Create assignment dict
    assignment = {
        "properties": {
            "displayName": "TestAssign",
            "policyDefinitionId": "/subscriptions/x/providers/Microsoft.Authorization/policySetDefinitions/init1",
        }
    }

    res = engine.create_assignment("myAssign", assignment, scope="/subscriptions/x")
    assert res.get("name") == "myAssign" or res.get("params_type") in (
        "DummyModel",
        "dict",
    )


def test_sdk_deploy_initiative_uses_model(monkeypatch, tmp_path):
    engine = AzurePolicyEngine(backend="sdk", policy_dir=str(tmp_path))
    dummy_client = types.SimpleNamespace()

    class DummySetDefs:
        def __init__(self):
            self.created = []

        def create_or_update(self, name, model):
            self.created.append((name, model))
            return types.SimpleNamespace(
                as_dict=lambda: {"name": name, "model_type": type(model).__name__}
            )

    dummy_client.policy_set_definitions = DummySetDefs()

    # Patch PolicyClient creation
    monkeypatch.setitem(
        __import__("sys").modules,
        "azure.mgmt.resource",
        types.SimpleNamespace(PolicyClient=lambda cred, sub: dummy_client),
    )
    # provide model class
    policy_module = types.SimpleNamespace()
    policy_module.models = types.SimpleNamespace(PolicySetDefinition=DummyModel)
    monkeypatch.setitem(
        __import__("sys").modules, "azure.mgmt.resource.policy", policy_module
    )

    initiative = {"properties": {"displayName": "MyInit", "policyDefinitions": []}}
    res = engine.deploy_initiative("init1", initiative)
    assert res.get("name") == "init1" or res.get("model_type") in ("DummyModel", "dict")



from __future__ import annotations

import argparse
import os
from typing import Optional

from .engine import AzurePolicyEngine
from . import cli_handlers


def _load_azure_credential() -> Optional[object]:
    try:
        from azure.identity import DefaultAzureCredential

        return DefaultAzureCredential()
    except Exception:
        return None


def main(argv=None):
    p = argparse.ArgumentParser(prog="azure-policy-engine")
    p.add_argument(
        "--backend",
        default=os.environ.get("AZPE_BACKEND", "file"),
        choices=["file", "sdk", "azcli"],
    )
    p.add_argument("--policies-dir", default=os.environ.get("AZPE_POLICY_DIR"))
    p.add_argument("--subscription-id", default=os.environ.get("AZPE_SUBSCRIPTION_ID"))
    sub = p.add_subparsers(dest="cmd")

    sub.add_parser("list")
    g = sub.add_parser("get")
    g.add_argument("name")
    d = sub.add_parser("deploy")
    d.add_argument("name")
    d.add_argument("file")
    sub.add_parser("assignments")
    ca = sub.add_parser("create-assignment")
    ca.add_argument("name")
    ca.add_argument("file")
    ua = sub.add_parser("update-assignment")
    ua.add_argument("name")
    ua.add_argument("file")
    da = sub.add_parser("delete-assignment")
    da.add_argument("name")
    sub.add_parser("compliance")
    a = sub.add_parser("add-assignment")
    a.add_argument("name")
    a.add_argument("file")
    ai = sub.add_parser("assign-initiative")
    ai.add_argument("initiative-name")
    ai.add_argument("assignment-name")
    ai.add_argument("file")
    i = sub.add_parser("list-initiatives")
    i.add_argument("--raw", action="store_true")
    p.add_argument(
        "--scope",
        default=None,
        help="Scope for assignments (e.g. /subscriptions/.. or /providers/Microsoft.Management/managementGroups/{id})",
    )
    p.add_argument(
        "--mgmt-group",
        dest="mgmt_group",
        default=None,
        help="Management group id (shorthand); builds scope /providers/Microsoft.Management/managementGroups/{id}",
    )
    p.add_argument(
        "--dry-run",
        dest="dry_run",
        action="store_true",
        help="Do not perform changes; print the effective payload and scope",
    )

    args = p.parse_args(argv)
    credential = _load_azure_credential() if args.backend == "sdk" else None
    engine = AzurePolicyEngine(
        backend=args.backend,
        policy_dir=args.policies_dir,
        subscription_id=args.subscription_id,
        credential=credential,
    )

    # Delegate command handling to the handlers module which keeps `main` simple.
    cli_handlers.handle_command(args, engine)


if __name__ == "__main__":
    main()



from __future__ import annotations

import json
import shlex
import subprocess
import tempfile
import os
from typing import Any, Dict, List, Optional


class AzCliError(RuntimeError):
    pass


class AzCliBackend:
    """Minimal wrapper around `az rest` to perform policy operations via the Azure CLI.

    This backend prefers `az rest` to call ARM endpoints because it allows passing
    arbitrary JSON payloads (safer for complex properties). It requires `az`
    being installed and authenticated.
    """

    def __init__(self, subscription_id: Optional[str] = None):
        self.subscription_id = subscription_id

    def _run_az(self, args: List[str]) -> Dict[str, Any]:
        cmd = ["az"] + args + ["--output", "json"]
        try:
            proc = subprocess.run(
                cmd,
                check=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
            )
            out = proc.stdout.strip()
            if not out:
                return {}
            try:
                return json.loads(out)
            except Exception:
                return {"raw": out}
        except subprocess.CalledProcessError as ex:
            raise AzCliError(
                f"az command failed: {ex.stderr.strip() or ex.stdout.strip()}"
            )

    def _az_rest(self, method: str, uri: str, body: Optional[Dict] = None) -> Any:
        args = ["rest", "--method", method, "--uri", uri]
        if body is not None:
            # write temp file and pass with --body @file. Ensure the temp file is removed after use.
            tf = tempfile.NamedTemporaryFile(
                "w+", suffix=".json", delete=False, encoding="utf-8"
            )
            try:
                try:
                    json.dump(body, tf, indent=2)
                    tf.flush()
                    tf_name = tf.name
                finally:
                    tf.close()

                args += ["--body", f"@{tf_name}"]
                result = self._run_az(args)
                return result
            finally:
                # attempt to remove the temporary file regardless of success or failure
                try:
                    if tf and os.path.exists(tf_name):
                        os.remove(tf_name)
                except Exception:
                    # best-effort cleanup; do not mask original exceptions
                    pass
        return self._run_az(args)

    # Policy definitions (list/get/create)
    def list_policy_definitions(self) -> List[Dict[str, Any]]:
        # Try subscription-scoped list when subscription provided, otherwise tenant-scoped
        if self.subscription_id:
            uri = f"https://management.azure.com/subscriptions/{self.subscription_id}/providers/Microsoft.Authorization/policyDefinitions?api-version=2021-06-01"
        else:
            uri = f"https://management.azure.com/providers/Microsoft.Authorization/policyDefinitions?api-version=2021-06-01"
        res = self._az_rest("GET", uri)
        # az rest returns dict with 'value' for list responses
        if isinstance(res, dict) and "value" in res:
            return res.get("value", [])
        if isinstance(res, list):
            return res
        return []

    def get_policy_definition(self, name: str) -> Dict[str, Any]:
        if self.subscription_id:
            uri = f"https://management.azure.com/subscriptions/{self.subscription_id}/providers/Microsoft.Authorization/policyDefinitions/{name}?api-version=2021-06-01"
        else:
            uri = f"https://management.azure.com/providers/Microsoft.Authorization/policyDefinitions/{name}?api-version=2021-06-01"
        return self._az_rest("GET", uri)

    def create_or_update_policy_definition(
        self, name: str, policy: Dict[str, Any]
    ) -> Dict[str, Any]:
        if self.subscription_id:
            uri = f"https://management.azure.com/subscriptions/{self.subscription_id}/providers/Microsoft.Authorization/policyDefinitions/{name}?api-version=2021-06-01"
        else:
            uri = f"https://management.azure.com/providers/Microsoft.Authorization/policyDefinitions/{name}?api-version=2021-06-01"
        return self._az_rest("PUT", uri, body=policy)

    # Assignments
    def list_assignments(self, scope: Optional[str] = None) -> List[Dict[str, Any]]:
        scope_path = scope or (
            f"/subscriptions/{self.subscription_id}" if self.subscription_id else ""
        )
        scope_path = scope_path.rstrip("/")
        if not scope_path:
            # list at tenant scope
            uri = f"https://management.azure.com/providers/Microsoft.Authorization/policyAssignments?api-version=2021-06-01"
        else:
            uri = f"https://management.azure.com{scope_path}/providers/Microsoft.Authorization/policyAssignments?api-version=2021-06-01"
        res = self._az_rest("GET", uri)
        if isinstance(res, dict) and "value" in res:
            return res.get("value", [])
        if isinstance(res, list):
            return res
        return []

    def create_or_update_assignment(
        self, name: str, assignment: Dict[str, Any], scope: Optional[str] = None
    ) -> Dict[str, Any]:
        scope_path = scope or (
            f"/subscriptions/{self.subscription_id}" if self.subscription_id else ""
        )
        if not scope_path:
            raise AzCliError(
                "scope must be provided when subscription id is not configured"
            )
        uri = f"https://management.azure.com{scope_path}/providers/Microsoft.Authorization/policyAssignments/{name}?api-version=2021-06-01"
        return self._az_rest("PUT", uri, body=assignment)

    def delete_assignment(
        self, name: str, scope: Optional[str] = None
    ) -> Dict[str, Any]:
        scope_path = scope or (
            f"/subscriptions/{self.subscription_id}" if self.subscription_id else ""
        )
        if not scope_path:
            raise AzCliError(
                "scope must be provided when subscription id is not configured"
            )
        uri = f"https://management.azure.com{scope_path}/providers/Microsoft.Authorization/policyAssignments/{name}?api-version=2021-06-01"
        return self._az_rest("DELETE", uri)

    # Initiatives (policy set definitions)
    def list_initiatives(self) -> List[Dict[str, Any]]:
        if self.subscription_id:
            uri = f"https://management.azure.com/subscriptions/{self.subscription_id}/providers/Microsoft.Authorization/policySetDefinitions?api-version=2021-06-01"
        else:
            uri = f"https://management.azure.com/providers/Microsoft.Authorization/policySetDefinitions?api-version=2021-06-01"
        res = self._az_rest("GET", uri)
        if isinstance(res, dict) and "value" in res:
            return res.get("value", [])
        if isinstance(res, list):
            return res
        return []

    def get_initiative(self, name: str) -> Dict[str, Any]:
        if self.subscription_id:
            uri = f"https://management.azure.com/subscriptions/{self.subscription_id}/providers/Microsoft.Authorization/policySetDefinitions/{name}?api-version=2021-06-01"
        else:
            uri = f"https://management.azure.com/providers/Microsoft.Authorization/policySetDefinitions/{name}?api-version=2021-06-01"
        return self._az_rest("GET", uri)

    def create_or_update_initiative(
        self, name: str, initiative: Dict[str, Any]
    ) -> Dict[str, Any]:
        if self.subscription_id:
            uri = f"https://management.azure.com/subscriptions/{self.subscription_id}/providers/Microsoft.Authorization/policySetDefinitions/{name}?api-version=2021-06-01"
        else:
            uri = f"https://management.azure.com/providers/Microsoft.Authorization/policySetDefinitions/{name}?api-version=2021-06-01"
        return self._az_rest("PUT", uri, body=initiative)
